% Given an infinitesimal generator Q, drift matrix D, this
% program computes the probability that the buffer length exceeds 
% a certain value x for a first order fluid flow model (FOFFM) 
% with finite buffer size B.
% Pr (Q > x) = a + f1 exp(G1 x) h1 + f2 exp(-G2(B-x)) h2 
% The model is obtained via multiplexing of K 3-state sources
% rho, N, T_on, C_onsquared are input to the system



util = 0.95;
K=10;
T_on = 2;
C_onsquared = 4*4;
source_rho = 0.4;
r = 1;

% construct the single source parameters

% from Tijms
p_1 = 0.5*(1 + sqrt((C_onsquared-1)/(C_onsquared+1)));
p_2 = 1-p_1;
mu_1 = 2*p_1 / T_on;
mu_2 = 2*p_2 / T_on;

T_off = (1 - source_rho)*T_on / source_rho;

r_10 = mu_1;
r_20 = mu_2;
r_01 = p_1/T_off;
r_02 = p_2/T_off;


C = K*source_rho*r/ util;
B = 10;
fix = 1;




S_minus = 0;
S_plus = 0;
S =(K+1)*(K+2)/2;
N = S-1;
A = zeros(N+1,N+1);
E = zeros(N+1,N+1);
E_=E; 
for n1=0:K
   for n2=0:K-n1
      
      temp = n1*(K+1) - n1*(n1-1)/2 + (n2+1);
      
      E(temp,temp) = (n1+n2)*r - C;
      E_(temp,temp)=1/E(temp,temp);
      if(E(temp,temp)>0) 
         S_plus = S_plus + 1;
         
      else S_minus =S_minus + 1;
      end
      
      
      temp1 = n1*(K+1) - n1*(n1-1)/2 + (n2+2);
      temp2 = (n1+1)*(K+1) - (n1+1)*(n1)/2 + (n2+1);
      temp3 = n1*(K+1) - n1*(n1-1)/2 + (n2);
      temp4 = (n1-1)*(K+1) - (n1-1)*(n1-2)/2 + (n2+1);
         
      
      
      if(n1==0 & n2==0)
         A(temp,temp1) = (K-n1-n2)*r_02;
         A(temp,temp2) = (K-n1-n2)*r_01;
         % A(temp,temp3) =  n2*r_20;
         % A(temp,temp4) = n1*r_10;
         A(temp,temp) = -(A(temp,temp1) + A(temp,temp2));
      elseif(n1==0 & n2==K)
         % A(temp,temp1) = (K-n1-n2)*r_02;
         % A(temp,temp2) = (K-n1-n2)*r_01;
         A(temp,temp3) =  n2*r_20;
         % A(temp,temp4) = n1*r_10;
         A(temp,temp) = -(A(temp,temp3));
      elseif(n1==0 & 0<n2<K )
         A(temp,temp1) = (K-n1-n2)*r_02;
         A(temp,temp2) = (K-n1-n2)*r_01;
         A(temp,temp3) =  n2*r_20;
         % A(temp,temp4) = n1*r_10;
         A(temp,temp) = -(A(temp,temp1) + A(temp,temp2) + A(temp,temp3));
      elseif(n1==K & n2==0)
         % A(temp,temp1) = (K-n1-n2)*r_02;
         %A(temp,temp2) = (K-n1-n2)*r_01;
         % A(temp,temp3) =  n2*r_20;
         A(temp,temp4) = n1*r_10;
         A(temp,temp) = -(A(temp,temp4));
      %elseif(n1==K & n2==K) 
         % A(temp,temp1) = (K-n1-n2)*r_02;
         % A(temp,temp2) = (K-n1-n2)*r_01;
         % A(temp,temp3) =  n2*r_20;
         % A(temp,temp4) = n1*r_10;
         % A(temp,temp) = -(A(temp,temp3) + A(temp,temp4));
      %elseif(n1==K & 0<n2<K)
         %  A(temp,temp1) = (K-n1-n2)*r_02;
         % A(temp,temp2) = (K-n1-n2)*r_01;
         % A(temp,temp3) =  n2*r_20;
         % A(temp,temp4) = n1*r_10;
         % A(temp,temp) = -(A(temp,temp1) + A(temp,temp3) + A(temp,temp4));
      elseif(0<n1<K & n2==0)
         A(temp,temp1) = (K-n1-n2)*r_02;
         A(temp,temp2) = (K-n1-n2)*r_01;
         % A(temp,temp3) =  n2*r_20;
         A(temp,temp4) = n1*r_10;
         A(temp,temp) = -(A(temp,temp1) + A(temp,temp2) + A(temp,temp4));
      %elseif(0<n1<K & n2==K)
         % A(temp,temp1) = (K-n1-n2)*r_02;
       %  A(temp,temp2) = (K-n1-n2)*r_01;
       %  A(temp,temp3) =  n2*r_20;
       %  A(temp,temp4) = n1*r_10;
        % A(temp,temp) = -(A(temp,temp2) + A(temp,temp3) + A(temp,temp4));
     else
        temp5 = 0;
         if(K ~= n1+n2) 
            A(temp,temp1) = (K-n1-n2)*r_02; 
            A(temp,temp2) = (K-n1-n2)*r_01; 
            temp5 = A(temp,temp1) + A(temp,temp2);
         end
         

         A(temp,temp3) =  n2*r_20;
         A(temp,temp4) = n1*r_10;
         A(temp,temp) = -(temp5 + A(temp,temp3) + A(temp,temp4));
         
         
      end
         end
end

      

T_bar = zeros(N+1,S_plus);
N_bar = zeros(N+1,S_minus);
overload = 0;
underload = 0;
for n1=0:K
   for n2=0:K-n1
      temp = n1*(K+1) - n1*(n1-1)/2 + (n2+1);
       if(E(temp,temp)>0)
          overload = overload +1;
          T_bar(temp,overload) = 1;     
       else 
           underload = underload +1;
           N_bar(temp,underload)=1;
        end
     end
  end


     
% A x_null = 0;
e = ones(N+1,1);
x_null = e;
x_null = x_null / (N+1);


% y_null A = 0;
y_null = null(A')';
y_null = y_null / sum(y_null);



% find the average drift d

d = y_null* diag(E);

% M_stable: number of stable eigenvalues
% M_unstable: number of strictly unstable eigenvalues




%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
Q=A;
R=E;
R_=E_;

% queue capacity
%B = 100;
B = 1;

% initial buffer level > 0
a = 1e-12;

% time horizon > 0
h = 10^4;

% initial distribution
pi0 = [1,zeros(1,size(Q,1)-1)]/[ones(size(Q,1),1),Q(:,2:end)];
pi0(diag(R)<=0) = 0;
pi0 = pi0 / sum(pi0);

% horizon PH
% Erlang-L
L = 20;
T_h = -diag(L/h*ones(L,1)) + [zeros(L-1,1), diag(L/h*ones(L-1,1)); zeros(1,L)];
v = [1, zeros(1,L-1)];
T0 = -T_h * ones(L,1); %%% -sum(T,2);


a = 1;
L = 3;
fi = -3.47863;
om = 1.03593;
T_h = [-a, 0,0; 0, -a, -om; 0, om, -a];
u = a*(a^2+om^2)/(a^2+om^2+a^2*cos(fi)-a*om*sin(fi));
g = u / a;
c = (a^2*cos(fi)-a*om*sin(fi)) / 2 / (a^2 + om^2 + a^2*cos(fi) - a*om*sin(fi));
d = (a*om*cos(fi)+a^2*sin(fi)) / 2 / (a^2 + om^2 + a^2*cos(fi) - a*om*sin(fi));
v = [g, c+d, c-d];
m1 = sum(v*inv(-T_h));
T_h = T_h * m1 / h;
T0 = -T_h * ones(L,1);


% horizon PH
% pade(p,q)
% L=10;
% [T_h,T0,v,d]=pade_approx(9,10,h);

% horizon ME
% L=11;
% [MESystem,cv] = TimeDelayME(L,h);
% T_h=MESystem.A;
% T0=MESystem.b;
% v=MESystem.c;



% find cdf at
xvec = [0 10 20 30 40 50 60 70 80 90 100]'/100;
return 


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

% [U,V] = additive_decompose_inversefree(E,A,M_unstable);
% pause
% 
% 
% 
% E_after = inv(U)*E*V;
% A_after = inv(U)*A*V;
% 
% 
% 
% 
% % above inv(U) E,A V are all block diagonal with three blocks, first block 
% % with eignevalue at zero, secondd block stable eigenvalues, last block unstable eignevalues
% 
% 
% E_11 = E_after(2:M_stable+1,2:M_stable+1);
% A_11 = A_after(2:M_stable+1,2:M_stable+1);
% E_22 = E_after(M_stable+2:N+1,M_stable+2:N+1);
% A_22 = A_after(M_stable+2:N+1,M_stable+2:N+1);
% 
%    
% G1 = A_11 / E_11;
% G2 = A_22 / E_22;
% 
% U_ = inv(U);
% L1 = U_(1,:);
% L2 = U_(2:M_stable+1,:);
% L3 = U_(M_stable+2:N+1,:);
% 
% 
% L1_bar = L1*T_bar;
% L2_bar = L2*T_bar;
% L3_bar = L3*T_bar;
% 
% L1_tilde = L1*N_bar;
% L2_tilde = L2*N_bar;
% L3_tilde = L3*N_bar;
% 
% 
% 
% b = [zeros(1,S_plus) y_null*N_bar];
% Big = [L1_bar L1_tilde ; L2_bar expm(G1*B)*L2_tilde; expm(-G2*B)*L3_bar L3_tilde];
% 
% soln = b / Big;
% c = soln(1);
% 
% a = 1 - c*L1*e;
% f1 = soln(2:M_stable+1);
% f2 = soln(M_stable+2:N+1);
% 
% h1 = -L2*e;
% h2 = -L3*e;
% 
% 
% 
% 
% 
%  x = 0:fix:B;
%  X_survivor = a* ones(B/fix+1,1);
%  X_survivor(1) = X_survivor(1) + f1*h1;
%  for i=2:B/fix+1
%     X_survivor(i) = X_survivor(i) + f1*expm(G1*(i*fix-fix))*h1;
%  end
% 
%  X_survivor(B/fix+1) = X_survivor(B/fix+1) + f2*h2;
%  for i=1:B/fix
%     X_survivor(i) = X_survivor(i) + f2*expm(-G2*(B-(i-1)*fix))*h2;
%  end
% 
% 
% x = 0:fix:B;
% X_density = zeros(B/fix+1,1);
% % X_density(1) = X_density(1) - f1*G1*h1;
% for i=2:B/fix+1
%    X_density(i) = X_density(i) - f1*G1*expm(G1*(i*fix-fix))*h1;
% end
% 
% X_density(B/fix+1) = X_density(B/fix+1) - f2*G2*h2;
% for i=1:B/fix
%    X_density(i) = X_density(i) - f2*G2*expm(-G2*(B-(i-1)*fix))*h2;
% end
% 
% 
% % performance measures
% 
% 
% 
% a_ = c*L1;
% f1_ = f1;
% h1_ = L2;
% f2_ = f2;
% h2_ = L3;
% 
% % F(x) is in the form a_ + f1_ exp(G1 x) h1_ + f2_ exp(-G2(B-x)) h2_
% % whether it satisfies the boundary condition at zero
% 
% boundary_zero = a_ + f1_*h1_ + f2_ *expm(-G2*B) *h2_;
% norm1 = norm(boundary_zero*T_bar,1);
% 
% % whether it satisfies the boundary condition at B
% boundary_B = a_ + f1_ * expm(G1*B)* h1_ + f2_* h2_;
% norm2 = norm(boundary_B*N_bar-y_null*N_bar,1);
% 
% % whether it satisfies the differential equation at zero
% boundary_zero_derivative = f1_*G1*h1_ + f2_*G2*expm(-G2*B)*h2_;
% norm3 = norm(boundary_zero_derivative*E-boundary_zero*A,1);
% 
% 
% 
% % whether it satisfies the differential equation at B
% boundary_B_derivative = f1_*G1*expm(G1*B)*h1_ + f2_*G2*h2_;
% norm4 = norm(boundary_B_derivative*E-boundary_B*A,1);
% 
% 
% norm_final = norm([boundary_zero*T_bar boundary_B*N_bar-y_null*N_bar boundary_zero_derivative*E-boundary_zero*A boundary_B_derivative*E-boundary_B*A],1);
% 
% 
% 
% 
% 
% 
% 
% 
% 

